"""
Contains code for installing and removing SWAI applications with the .swai format.
"""

import os
import shutil
import yaml
import requests
import urllib.parse
import swai_inst.utils as utils
from swai_inst.exceptions import (
    SwaiFileError, SwaiParseError, SwaiDownloadError, SwaiInstallationError,
    SwaiDesktopFileError, SwaiValidationError, SwaiErrorCodes, create_swai_error_from_exception
)

def _parse_swai_file(swai_file_path, verbose=False) -> dict:
    """
    Reads and parses the content of the .swai file as YAML.

    Args:
        swai_file_path (str): The path to the .swai file.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.

    Returns:
        dict: The parsed YAML data.

    Raises:
        SwaiFileError: If the file cannot be read.
        SwaiParseError: If the file cannot be parsed as YAML.
    """
    try:
        with open(swai_file_path, 'r') as f:
            # Use safe_load to prevent potential security issues
            data = yaml.safe_load(f)
            if data is None:
                raise SwaiParseError(
                    f"SWAI file is empty: {swai_file_path}",
                    file_path=swai_file_path,
                    error_code=SwaiErrorCodes.INVALID_SWAI_FORMAT
                )
            return data
    except FileNotFoundError as e:
        raise SwaiFileError(
            f"SWAI file not found: {swai_file_path}",
            file_path=swai_file_path,
            error_code=SwaiErrorCodes.FILE_NOT_FOUND,
            original_exception=e
        )
    except yaml.YAMLError as e:
        raise SwaiParseError(
            f"Error parsing SWAI file '{swai_file_path}': {e}",
            file_path=swai_file_path,
            error_code=SwaiErrorCodes.YAML_PARSE_ERROR,
            original_exception=e
        )
    except Exception as e:
        raise create_swai_error_from_exception(
            e,
            f"Unexpected error reading SWAI file '{swai_file_path}'"
        )


def _download_icon(icon_url, icon_base_dir, app_id, verbose=False) -> str | None:
    """
    Downloads the icon from the given URL to the specified directory.

    Args:
        icon_url (str): The URL of the icon.
        icon_base_dir (str): The base directory to save the icon.
        app_id (str): The application ID (used for filename).
        verbose (bool, optional): Whether to print verbose output. Defaults to False.

    Returns:
        str: The full path to the downloaded icon file, or None if download fails or icon_url is None/empty.
        None: if download fails or icon_url is None/empty.
    """
    if not icon_url:
        return None

    try:
        # Send a GET request to the icon URL
        response = requests.get(icon_url, stream=True, timeout=10)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx or 5xx)

        # Determine file extension from the URL
        parsed_url = urllib.parse.urlparse(icon_url)
        path = parsed_url.path
        ext = os.path.splitext(path)[1].lower()
        
        if not ext:
            # If no extension in URL, try to guess from content type
            content_type = response.headers.get('Content-Type', '').lower()
            if 'png' in content_type:
                ext = '.png'
            elif 'jpeg' in content_type or 'jpg' in content_type:
                ext = '.jpg'
            elif 'svg' in content_type:
                ext = '.svg'
            else:
                if verbose:
                    print(f"Warning: Could not determine icon file extension from URL or Content-Type for '{icon_url}'. Skipping download.")
                return None

        # Sanitize app_id for filename and ensure it's not empty
        if not app_id:
            app_id = "unknown_app"
            if verbose:
                print("Warning: 'app_id' not found in SWAI file, using 'unknown_app' for icon filename.")

        # Create a safe filename
        icon_filename = f"swai.{app_id}{ext}"
        icon_file_path = os.path.join(icon_base_dir, icon_filename)

        # Ensure the directory exists
        os.makedirs(icon_base_dir, exist_ok=True)

        # Save the icon content to the file
        with open(icon_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:  # Filter out keep-alive chunks
                    f.write(chunk)

        if verbose:
            print(f"Successfully downloaded icon from '{icon_url}' to '{icon_file_path}'")
            
        return icon_file_path

    except requests.exceptions.RequestException as e:
        if verbose:
            print(f"Error downloading icon from '{icon_url}': {e}")
        return None
    except (IOError, OSError) as e:
        if verbose:
            print(f"Error saving downloaded icon to '{icon_file_path}': {e}")
        return None


def _generate_desktop_content(app_name, icon_path, swai_file_path, app_id, categories=None):
    """
    Generates the content for the .desktop file.

    Args:
        app_name (str): The name of the application.
        icon_path (str): The local path to the application icon file, or None.
        swai_file_path (str): The full path to the copied .swai file.
        app_id (str): The application ID for StartupWMClass.
        categories (list, optional): A list of application categories. Defaults to None.

    Returns:
        str: The content of the .desktop file.
    """
    content = f"""[Desktop Entry]
Version=1.0
Type=Application
Name={app_name}
Exec=/bin/swai "{swai_file_path}"
Terminal=false
StartupWMClass={app_id}
"""
    if icon_path:
        content += f"Icon={icon_path}\n" # Use the local icon path

    if categories:
        content += f"Categories={';'.join(categories)};\n"

    return content


def _write_desktop_file(desktop_dir, app_id, desktop_content, verbose=False):
    """
    Writes the generated .desktop file content to the destination directory.

    Args:
        desktop_dir (str): The destination directory for the .desktop file.
        app_id (str): The application ID (used for the filename).
        desktop_content (str): The content to write to the .desktop file.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.

    Returns:
        str: The full path to the created .desktop file.

    Raises:
        SwaiDesktopFileError: If the desktop file cannot be created.
    """
    if not app_id:
        app_id = "unknown_app"  # Fallback if app_id is missing
        if verbose:
            print("Warning: Using 'unknown_app' for desktop filename due to missing app_id")

    # Create a safe filename
    desktop_filename = f"swai.{app_id}.desktop"
    desktop_file_path = os.path.join(desktop_dir, desktop_filename)

    try:
        # Ensure the directory exists
        os.makedirs(desktop_dir, exist_ok=True)
        
        # Write the desktop file
        with open(desktop_file_path, 'w', encoding='utf-8') as f:
            f.write(desktop_content)

        # Make the desktop file executable
        os.chmod(desktop_file_path, 0o755)
        
        if verbose:
            print(f"Successfully created desktop file at '{desktop_file_path}'")
            
        return desktop_file_path
        
    except (IOError, OSError) as e:
        raise SwaiDesktopFileError(
            f"Error writing desktop file '{desktop_file_path}': {e}",
            desktop_file_path=desktop_file_path,
            error_code=SwaiErrorCodes.DESKTOP_FILE_ERROR,
            original_exception=e
        )


def install_from_str(swai_content, local_install=False, verbose=False):
    """
    Installs a SWAI application from a string containing the .swai file content.

    Args:
        swai_content (str): The content of the .swai file.
        local_install (bool, optional): Whether to install the application locally (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    
    Returns:
        tuple: (app_name, app_id) of the installed application
        
    Raises:
        SwaiParseError: If the SWAI content cannot be parsed.
        SwaiValidationError: If required fields are missing from the SWAI content.
        SwaiInstallationError: If there are issues during the installation process.
    """
    # Parse the YAML content
    try:
        swai_data = yaml.safe_load(swai_content)
    except yaml.YAMLError as e:
        raise SwaiParseError(
            f"Invalid YAML in SWAI content: {e}",
            error_code=SwaiErrorCodes.YAML_PARSE_ERROR,
            original_exception=e
        )

    # Extract app information
    try:
        app_name = swai_data['app_name']
        app_id = swai_data['app_id']
        icon_url = swai_data['icon_url']
        categories = swai_data['categories']

    except KeyError as e:
        raise SwaiValidationError(
            f"Missing required field in SWAI content: {e}",
            field_name=str(e),
            error_code=SwaiErrorCodes.MISSING_REQUIRED_FIELD,
            original_exception=e
        )

    # Get installation paths
    swai_dir, desktop_dir, icon_base_dir = utils.get_install_paths(local_install=local_install)
    
    # Create the SWAI directory if it doesn't exist
    os.makedirs(swai_dir, exist_ok=True)

    # Create a filename based on app_id
    swai_filename = f"{app_id}.swai"
    installed_swai_path = os.path.join(swai_dir, swai_filename)

    # Write the .swai file to the installation directory
    try:
        with open(installed_swai_path, 'w') as f:
            f.write(swai_content)
        if verbose:
            print(f"Installed SWAI file to {installed_swai_path}")
    except IOError as e:
        raise SwaiInstallationError(
            f"Error writing SWAI file to {installed_swai_path}: {e}",
            app_id=app_id,
            installation_step='swai_file_creation',
            error_code=SwaiErrorCodes.FILE_WRITE_ERROR,
            original_exception=e
        )

    # Download the icon if it exists
    downloaded_icon_path = None
    if icon_url:
        downloaded_icon_path = _download_icon(icon_url, icon_base_dir, app_id, verbose=verbose)
        if downloaded_icon_path is None and verbose:
            print(f"Warning: Failed to download icon from '{icon_url}'. The desktop entry may not display an icon.")

    # Generate the .desktop file content
    desktop_content = _generate_desktop_content(
        app_name,
        downloaded_icon_path,
        installed_swai_path,
        app_id,
        categories
    )

    try:
        # Write the desktop file using app_id for the filename
        desktop_file_path = _write_desktop_file(desktop_dir, app_id, desktop_content, verbose=verbose)
    except SwaiDesktopFileError as e:
        # Clean up the installed file if writing desktop file failed
        if os.path.exists(installed_swai_path):
            try:
                os.remove(installed_swai_path)
                if verbose:
                    print(f"Cleaned up installed file '{installed_swai_path}' due to desktop file writing error.")
            except IOError as cleanup_e:
                if verbose:
                    print(f"Error cleaning up installed file '{installed_swai_path}': {cleanup_e}")
        # Clean up downloaded icon if it exists
        if downloaded_icon_path and os.path.exists(downloaded_icon_path):
            try:
                os.remove(downloaded_icon_path)
                if verbose:
                    print(f"Cleaned up downloaded icon file '{downloaded_icon_path}' due to desktop file writing error.")
            except IOError as cleanup_e:
                if verbose:
                    print(f"Error cleaning up downloaded icon file '{downloaded_icon_path}': {cleanup_e}")
        raise SwaiInstallationError(
            f"Failed to write desktop file: {e}",
            app_id=app_id,
            installation_step='desktop_file_creation',
            error_code=SwaiErrorCodes.INSTALLATION_FAILED,
            original_exception=e
        )

    if verbose:
        print(f"Successfully installed '{app_name}' to '{desktop_file_path}'")
    
    return (app_name, app_id)


def install_from_file(swai_file_path, local_install=False, verbose=False):
    """
    Installs a SWAI application from a local .swai file.

    Args:
        swai_file_path (str): The path to the .swai file to install.
        local_install (bool, optional): Whether to install the application locally (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    
    Returns:
        tuple: (app_name, app_id) of the installed application
        
    Raises:
        SwaiFileError: If the SWAI file cannot be found or read.
    """
    # Check if the file exists
    if not os.path.isfile(swai_file_path):
        raise SwaiFileError(
            f"SWAI file not found: {swai_file_path}",
            file_path=swai_file_path,
            error_code=SwaiErrorCodes.FILE_NOT_FOUND
        )

    # Read the .swai file
    try:
        with open(swai_file_path, 'r') as f:
            swai_content = f.read()
    except IOError as e:
        raise SwaiFileError(
            f"Error reading SWAI file '{swai_file_path}': {e}",
            file_path=swai_file_path,
            error_code=SwaiErrorCodes.FILE_READ_ERROR,
            original_exception=e
        )
    
    # Install from the file content
    return install_from_str(swai_content, local_install=local_install, verbose=verbose)


def install_from_url(url, local_install=False, verbose=False):
    """
    Installs a SWAI application from a URL.

    Args:
        url (str): The URL of the .swai file to install.
        local_install (bool, optional): Whether to install the application locally (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    
    Returns:
        tuple: (app_name, app_id) of the installed application
        
    Raises:
        SwaiDownloadError: If there are issues downloading the file from the URL.
    """
    # Install from the downloaded content
    return install_from_str(utils.download_swai_to_string(url), local_install=local_install, verbose=verbose)


def remove_from_file(swai_file_path, local_install=False, verbose=False, remove_data=True):
    """
    Removes a SWAI application from a .swai file.

    Args:
        swai_file_path (str): The path to the .swai file to remove.
        local_install (bool, optional): Whether to remove the application locally (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    """
    
    swai_dir, desktop_dir, icon_base_dir = utils.get_install_paths(local_install)
    
    # Parse the .swai file
    try:
        swai_data = _parse_swai_file(swai_file_path, verbose=verbose)
    except Exception as e:
        raise Exception(f"Error parsing SWAI file '{swai_file_path}': {e}")
    
    # Extract required information
    app_name = swai_data.get('app_name')
    app_id = swai_data.get('app_id')
    icon_url = swai_data.get('icon_url')
    categories = swai_data.get('categories')
    
    # Remove the .swai file
    try:
        os.remove(swai_file_path)
    except Exception as e:
        if verbose:
            print(f"Warning: Failed to remove .swai file '{swai_file_path}': {e}")
        
        # If the SWAI file is not found, it's already removed, so continue
        # Otherwise, re-raise the exception
        if not isinstance(e, FileNotFoundError):
            raise

    # Remove the desktop file
    desktop_file_path = os.path.join(desktop_dir, f"swai.{app_id}.desktop")
    try:
        os.remove(desktop_file_path)        
    except FileNotFoundError:
        # Desktop file not found, likely already removed or never created, ignore
        pass
    except Exception as e:
        if verbose:
            print(f"Warning: Failed to remove desktop file '{desktop_file_path}': {e}")
            # Re-raise other exceptions
            raise

    # Remove the icon if it exists
    if icon_url:
        try:
            # Determine file extension from the URL, similar to _download_icon
            parsed_url = urllib.parse.urlparse(icon_url)
            path = parsed_url.path
            ext = os.path.splitext(path)[1].lower()
            
            if not ext:
                 # If no extension in URL, try to guess from content type (requires downloading/checking headers - skipping for simplicity in removal)
                 # For removal, we'll just try a common extension like .png or .svg
                 # A more robust solution would involve storing the actual installed icon path in the .swai file.
                 # For now, we'll attempt removal with common extensions or rely on the swai.app_id.* pattern.
                 # Let's try to glob for swai.app_id.*
                 import glob
                 icon_pattern = os.path.join(icon_base_dir, f"swai.{app_id}.*")
                 matching_icons = glob.glob(icon_pattern)
                 for icon_to_remove in matching_icons:
                     try:
                         os.remove(icon_to_remove)
                         if verbose:
                             print(f"Successfully removed icon file '{icon_to_remove}'")
                     except Exception as e:
                         if verbose:
                             print(f"Warning: Failed to remove icon file '{icon_to_remove}': {e}")
                 # Return after attempting glob removal
                 # We don't want to try specific extensions if we globbed
                 return # Exit the icon removal block

            icon_path = os.path.join(icon_base_dir, f"swai.{app_id}{ext}")
            os.remove(icon_path)
            if verbose:
                 print(f"Successfully removed icon file '{icon_path}'")
        except FileNotFoundError:
            # Icon file not found, likely already removed or never created, ignore
            pass
        except Exception as e:
            if verbose:
                print(f"Warning: Failed to remove icon file for app ID '{app_id}': {e}")
            # Re-raise other exceptions
            raise

    # Remove data directory if remove_data is True and it exists
    if remove_data:
        # Get data path
        data_path = os.path.join(utils.get_install_paths(local_install)[0], "..", "app_data", app_id)
        data_path = os.path.abspath(data_path) # Resolve potential '..'

        if os.path.isdir(data_path) and os.access(data_path, os.W_OK | os.X_OK):
             # Check if this is a system-wide install and if a local install exists
             if not local_install:
                 local_swai_share = os.path.join(os.path.expanduser("~"), ".local", "share", "swai")
                 local_app_path = os.path.join(local_swai_share, "apps", f"{app_id}.swai")
                 if os.path.exists(local_app_path):
                     if verbose:
                         print(f"Skipping data removal for '{app_id}' as a local installation exists.")
                     return # Do not remove data if a local install exists

             # If we reach here, either it's a local install or a system-wide install with no conflicting local install
             try:
                 shutil.rmtree(data_path)
                 if verbose:
                     print(f"Successfully removed data directory '{data_path}' for app '{app_id}'.")
             except Exception as e:
                 if verbose:
                     print(f"Warning: Failed to remove data directory '{data_path}' for app '{app_id}': {e}")
                 # Re-raise other exceptions
                 raise
        elif verbose:
            if not os.path.isdir(data_path):
                 print(f"No data directory found at '{data_path}' for app '{app_id}'.")
            else:
                 print(f"Data directory '{data_path}' for app '{app_id}' is not writable or executable by the current user. Skipping removal.")

    if verbose and app_name:
        print(f"Successfully removed '{app_name}' (ID: {app_id}).")
    elif verbose:
        print(f"Successfully removed app with ID: {app_id}.")


def remove_from_app_id(app_id, local_install=False, verbose=False,  remove_data=True):
    """
    Removes a SWAI application by its app_id.

    Args:
        app_id (str): The app_id of the application to remove.
        local_install (bool, optional): Whether to remove the application locally (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    """

    swai_dir, _, _ = utils.get_install_paths(local_install)
    
    # Get the .swai file path
    swai_file_path = os.path.join(swai_dir, f"{app_id}.swai")
    
    # Remove the .swai file
    remove_from_file(swai_file_path, local_install=local_install, verbose=verbose, remove_data=remove_data)

