import os
import shutil
import urllib

import requests
import yaml

from typing import List, Tuple
from swai_inst.exceptions import SwaiDownloadError, SwaiErrorCodes

def get_install_paths(local_install=False) -> Tuple[str, str, str]:
    """
    Returns the paths for the SWAI directory, desktop directory, and icon base directory.

    Args:
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.

    Returns:
        tuple: A tuple containing the paths for the SWAI directory, desktop directory, and icon base directory.
    """
    # System-wide installation paths
    swai_dir = '/usr/local/share/swai/apps'
    desktop_dir = '/usr/local/share/applications'
    icon_base_dir = '/usr/local/share/icons/hicolor/scalable/apps'  # Standard system icon path

    if local_install:
        # User-local installation paths
        home_dir = os.path.expanduser('~')
        swai_dir = os.path.join(home_dir, '.local', 'share', 'swai', 'apps')
        desktop_dir = os.path.join(home_dir, '.local', 'share', 'applications')
        icon_base_dir = os.path.join(home_dir, '.local', 'share', 'icons', 'hicolor', 'scalable', 'apps') # Standard user icon path

    # Ensure the icon base directories exists
    os.makedirs(icon_base_dir, exist_ok=True)
    os.makedirs(desktop_dir, exist_ok=True)
    os.makedirs(swai_dir, exist_ok=True)

    return swai_dir, desktop_dir, icon_base_dir


def get_installed_app_ids(local_install=False) -> List[str]:
    """
    Returns a list of installed app IDs.

    Args:
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.

    Returns:
        list: A list of installed app IDs.
    """
    swai_dir, _, _ = get_install_paths(local_install)
    app_ids = []
    
    for file_name in os.listdir(swai_dir):
        if file_name.endswith('.swai'):
            app_id = file_name[:-5]  # Remove the '.swai' extension
            app_ids.append(app_id)
    
    return app_ids


def get_installed_file_paths(app_id, local_install=False) -> List[str]:
    """Returns a list of file paths for the given app ID."""
    
    swai_dir, _, _ = get_install_paths(local_install)
    
    file_paths = []
    
    for file_name in os.listdir(swai_dir):
        if file_name.startswith(app_id):
            file_paths.append(os.path.join(swai_dir, file_name))
    
    return file_paths
    

def download_swai_to_string(url) -> str:
    """
    Downloads the .swai file from the given URL and returns its content as a string.

    Args:
        url (str): The URL of the .swai file.

    Returns:
        str: The content of the .swai file.
        
    Raises:
        SwaiDownloadError: If there are issues downloading the file.
    """
    try:
        response = requests.get(url)
        response.raise_for_status()
        return response.text
    except requests.exceptions.RequestException as e:
        raise SwaiDownloadError(
            f"Error downloading .swai file from '{url}': {e}",
            url=url,
            error_code=SwaiErrorCodes.DOWNLOAD_FAILED,
            original_exception=e
        )
